// Copyright (c) 2016, root ext. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

part of dna;

class DynamicLibrary<T> {
  final int STACK_SIZE = 256;

  Type type;
  String name;
  int libraryPointer;

  DynamicLibrary(){
    this.type = T;
    //print('type $type');

    InstanceMirror im = reflectClass(type).metadata.first.getField(#name);
    this.name = im.reflectee;

    libraryPointer = nLoadLibrary(this.name);
    //print('lib pointer $libraryPointer');
  }

  dynamic noSuchMethod(Invocation invocation){
    //print('method ${MirrorSystem.getName(invocation.memberName)}');

    ClassMirror classMirror = reflectClass(type);
    MethodMirror methodMirror = classMirror.instanceMembers.values.where((e)=> e.isRegularMethod && e.owner == classMirror && e.simpleName == invocation.memberName).first;

    List<TypedData> data = new List<TypedData>(invocation.positionalArguments.length);

    var methodPointer = nFindSymbol(libraryPointer, MirrorSystem.getName(invocation.memberName));
    //print('method pointer $methodPointer');
    var vm = nCreateCallVM(STACK_SIZE);

    before(invocation, methodMirror, vm, data);

    var result = call(methodMirror, vm, methodPointer);

    after(invocation, methodMirror, data);

    nFreeCallVM(vm);
    return result;
  }

  void before(Invocation invocation, MethodMirror methodMirror, int vm, List<TypedData> data){
    int index = 0;
    var params = methodMirror.parameters;
    List<Object> args = new List<Object>();
    List<String> signature = new List<String>();

    for (var param in params){
      //print('param ${MirrorSystem.getName(param.simpleName)}');
      var annot = Util.paramAnnot(param, Arg);
      var argType = annot != null ? annot.getField(#type).reflectee : typeToArgType(param.type);
      var arg = invocation.positionalArguments[index];

      signature.add(toSig[argType]);
      args.add(arg);

      switch (argType){
        case ARG_POINTER: {
          data[index] = new Uint64List(1);
          signature[signature.length-1] = toSig[ARG_TYPEDDATA];
          args[args.length-1] = data[index];
        }
        break;
        case ARG_STRING: {
          String str = arg.value;
          data[index] = new Uint8List(str.length+1);
          signature[signature.length-1] = toSig[ARG_TYPEDDATA];
        }
        break;
        case ARG_LISTINT: {
          if (arg != null) {
            List<int> list = arg.value;
            data[index] = new Uint32List(list.length);
            args[args.length-1] = data[index];
            signature[signature.length-1] = toSig[ARG_TYPEDDATA];
          }
        }
        break;
      }

      index++;
    }

    nArgs(vm, signature.join(''), args);
  }

  void after(Invocation invocation, MethodMirror methodMirror, List<TypedData> data){
    int index = 0;
    for (var param in methodMirror.parameters) {
      var annot = Util.paramAnnot(param, Arg);
      var argType = annot != null ? annot.getField(#type).reflectee : typeToArgType(param.type);

      switch (argType) {
        case ARG_POINTER: {
          Uint64List list = data[index];
          invocation.positionalArguments[index].value = list[0];
        }
        break;
        case ARG_STRING: {
          Uint8List list = data[index];
          invocation.positionalArguments[index].value = new String.fromCharCodes(list);
        }
        break;
        case ARG_LISTINT: {
          var arg = invocation.positionalArguments[index];
          if (arg != null) {
            Uint32List t = data[index];
            arg.value = t.toList();
          }
        }
        break;
      }

      index++;
    }
  }

  dynamic call(MethodMirror methodMirror, int vm, int methodPointer){

    var annot = Util.methodAnnot(methodMirror, Ret);
    var retType = annot != null ? annot.getField(#type).reflectee : typeToArgType(methodMirror.returnType);

    var result;
    if (retType == ARG_VOID){
      nCall(vm, methodPointer, toSig[retType]);
    } else {
      result = nCall(vm, methodPointer, toSig[retType]);
    }

    return result;
  }

  int typeToArgType(Object type){
    if (type == reflectClass(int)){
      return ARG_INT;
    } else if (type == reflectClass(double)){
      return ARG_DOUBLE;
    } else if (type == reflectClass(bool)){
      return ARG_BOOL;
    } else if (type == reflectClass(String)){
      return ARG_STRING;
    } else {
      return ARG_VOID;
    }
  }
}

class Strings {
  static String fill(String char, int length){
    return new String.fromCharCodes(new List<int>.filled(length, char.codeUnitAt(0)));
  }
}

class Util {
  static dynamic methodAnnot(MethodMirror m, Type t){
    var annots = m.metadata.where((e)=> e.type == reflectClass(t)).toList();
    return annots.isNotEmpty ? annots[0] : null;
  }

  static dynamic paramAnnot(ParameterMirror m, Type t){
    var annots = m.metadata.where((e)=> e.type == reflectClass(t)).toList();
    return annots.isNotEmpty ? annots[0] : null;
  }
}

const int ARG_VOID = -1;
const int ARG_CHAR = 0;
const int ARG_SHORT = 1;
const int ARG_INT = 2;
const int ARG_LONG = 3;
const int ARG_LONGLONG = 4;
const int ARG_FLOAT = 5;
const int ARG_DOUBLE= 6;
const int ARG_BOOL = 7;
const int ARG_POINTER = 8;
const int ARG_STRUCT = 9;
const int ARG_STRING = 10;
const int ARG_TYPEDDATA = 11;
const int ARG_LISTINT = 12;
const int ARG_LISTSTRING = 13;

final Map<int, String> toSig = {
  ARG_VOID : 'V',
  ARG_CHAR : 'c',
  ARG_SHORT : 's',
  ARG_INT : 'i',
  ARG_LONG : 'j',
  ARG_LONGLONG : 'l',
  ARG_FLOAT : 'f',
  ARG_DOUBLE : 'd',
  ARG_BOOL : 'B',
  ARG_POINTER : 'p',
  ARG_STRUCT : 'T',

  ARG_STRING : 'Z',
  ARG_TYPEDDATA : 'D',
  ARG_LISTINT : 'X',
  ARG_LISTSTRING : 'Y',
};

class Arg {
  final int type;

  const Arg(this.type);
}

class Ret {
  final int type;

  const Ret(this.type);
}

class Lib {
  final String name;

  const Lib(this.name);
}

class Ref<T> {
  T value;

  Ref(this.value);
}

class Struct {

  const Struct();
}

class Field {
  final int type;
  final int index;

  const Field(this.type, this.index);
}
